/*
 * Asterisk -- A telephony toolkit for Linux.
 *
 * Routines implementing music on hold
 * 
 * Copyright (C) 1999, Mark Spencer
 *
 * Mark Spencer <markster@linux-support.net>
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License
 */

#include <asterisk/lock.h>
#include <asterisk/file.h>
#include <asterisk/logger.h>
#include <asterisk/channel.h>
#include <asterisk/pbx.h>
#include <asterisk/options.h>
#include <asterisk/module.h>
#include <asterisk/translate.h>
#include <asterisk/say.h>
#include <asterisk/channel_pvt.h>
#include <asterisk/musiconhold.h>
#include <asterisk/config.h>
#include <asterisk/utils.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/time.h>
#include <sys/signal.h>
#include <netinet/in.h>
#include <sys/stat.h>
#include <dirent.h>
#ifdef ZAPATA_MOH
#ifdef __linux__
#include <linux/zaptel.h>
#else
#include <zaptel.h>
#endif /* __linux__ */
#endif
#include <unistd.h>
#include <sys/ioctl.h>

#define EXTRA_LOG 1
#define ELEMOF(a) (sizeof(a)/sizeof(a[0]))

#ifdef POSTGRES_MOH
#define SQL_MOH 1
#include <postgresql/libpq-fe.h>
#endif /* POSTGRES_MOH */

#ifdef MYSQL_MOH
#define SQL_MOH 1
#include <mysql/mysql.h>
#error "MYSQL support is not implemented yet"
#endif /* MYSQL_MOH */

static char *app0 = "MusicOnHold";
static char *app1 = "WaitMusicOnHold";
static char *app2 = "SetMusicOnHold";

static char *synopsis0 = "Play Music On Hold indefinitely";
static char *synopsis1 = "Wait, playing Music On Hold";
static char *synopsis2 = "Set default Music On Hold class";

static char *descrip0 = "MusicOnHold(class): "
"Plays hold music specified by class.  If omitted, the default\n"
"music source for the channel will be used. Set the default \n"
"class with the SetMusicOnHold() application.\n"
"Returns -1 on hangup.\n"
"Never returns otherwise.\n";

static char *descrip1 = "WaitMusicOnHold(delay): "
"Plays hold music specified number of seconds.  Returns 0 when\n"
"done, or -1 on hangup.  If no hold music is available, the delay will\n"
"still occur with no sound.\n";

static char *descrip2 = "SetMusicOnHold(class): "
"Sets the default class for music on hold for a given channel.  When\n"
"music on hold is activated, this class will be used to select which\n"
"music is played.\n";

struct mohclass {
	char class[80];
	char dir[256];
	char miscargs[256];
	int destroyme;
	int pid;		/* PID of mpg123 */
	int quiet;
	int single;
	int custom;
	pthread_t thread;
	struct mohdata *members;
	/* Source of audio */
	int srcfd;
#ifdef ZAPATA_MOH
	/* FD for timing source */
	int pseudofd;
#endif
	ast_mutex_t      lock;
#ifdef POSTGRES_MOH
	int              deadinsyncing;
	struct mohclass* new_mohclass;
#endif
	struct mohclass *next;
};

struct mohdata {
	int pipe[2];
	int origwfmt;
	struct mohclass *parent;
	struct mohdata *next;
};

static struct mohclass *mohclasses;

AST_MUTEX_DEFINE_STATIC(moh_lock);

#define LOCAL_MPG_123 "/usr/local/bin/mpg123"
#define MPG_123 "/usr/bin/mpg123"
#define MAX_MP3S 256

/*********************************************************************************
 * class moh_class (as much as you get get to a class in C :))
 ********************************************************************************/
static struct mohclass* mohclass_new( 
	const char *name, 
	const char *mode, 
	const char *dir, 
	const char *miscargs ) {
	struct mohclass* _this = NULL;
	if( !strcasecmp(mode, "mp3") || 
	    !strcasecmp(mode, "mp3nb") || 
	    !strcasecmp(mode, "quietmp3") || 
	    !strcasecmp(mode, "quietmp3nb") || 
	    !strcasecmp(mode, "httpmp3") || 
	    !strcasecmp(mode, "custom")) {
		_this = malloc(sizeof(struct mohclass));
		if( _this ) {
			memset(_this,0,sizeof(struct mohclass));
			strncpy(_this->class, name, sizeof(_this->class) - 1);
			if (miscargs)
				strncpy(_this->miscargs, miscargs, sizeof(_this->miscargs) - 1);
			if (!strcasecmp(mode, "custom"))
				_this->custom = 1;
			if (!strcasecmp(mode, "mp3nb") || !strcasecmp(mode, "quietmp3nb"))
				_this->single = 1;
			if (!strcasecmp(mode, "quietmp3") || !strcasecmp(mode, "quietmp3nb"))
				_this->quiet = 1;
			strncpy(_this->dir, dir, sizeof(_this->dir) - 1);
			_this->srcfd = -1;
#ifdef ZAPATA_MOH
			/* It's an MP3 Moh -- Open /dev/zap/pseudo for timing...  Is
			   there a better, yet reliable way to do this? */
			_this->pseudofd = open("/dev/zap/pseudo", O_RDONLY);
			if (_this->pseudofd < 0) {
				ast_log(LOG_WARNING, "Unable to open pseudo channel for timing...  Sound may be choppy.\n");
			} else {
				int x = 320;
				ioctl(_this->pseudofd, ZT_SET_BLOCKSIZE, &x);
			}
#endif
			ast_mutex_init(&_this->lock);
			return _this;
		} else {
			ast_log(LOG_WARNING,"Unable to allocate %d bytes for mohclass\n",sizeof(struct mohclass));
		}
	}
	else {
		ast_log(LOG_WARNING, "Don't know how to do a mode '%s' music on hold\n", mode);
	}
	return _this;
}

static struct mohclass* mohclass_find( struct mohclass* _this, const char* name ) {
	struct mohclass* tmp;
	for( tmp=_this; tmp; tmp=tmp->next ) {
		if( !strcasecmp(name,tmp->class) )
			return tmp;
	}
	return NULL;
}

static int mohclass_make_mp3_args( struct mohclass* _this, int argvsize, char* argv[] ) {

	int files;
	DIR* dir;
	int argc = 0;
	struct dirent *de;
	char xargs[256];
	char *argptr;
	char fns[MAX_MP3S][80];
	
	dir = opendir(_this->dir);
	if (!dir) {
		ast_log(LOG_WARNING, "%s is not a valid directory\n", _this->dir);
		return -1;
 	}

	if (!_this->custom) {
		argv[argc++] = "mpg123";
		argv[argc++] = "-q";
		argv[argc++] = "-s";
		argv[argc++] = "--mono";
		argv[argc++] = "-r";
		argv[argc++] = "8000";
		
		if (!_this->single) {
			argv[argc++] = "-b";
			argv[argc++] = "2048";
		}

		argv[argc++] = "-f";

		if (_this->quiet)
			argv[argc++] = "4096";
		else
			argv[argc++] = "8192";
	}

	/* Look for extra arguments and add them to the list */
	strncpy(xargs, _this->miscargs, sizeof(xargs) - 1);
	argptr = xargs;
	while(argptr && !ast_strlen_zero(argptr) && (argc<argvsize)) {
		argv[argc++] = argptr;
		argptr = strchr(argptr, ',');
		if (argptr) {
			*argptr = '\0';
			argptr++;
		}
	}
	
	files = 0;
	while((de = readdir(dir)) && (argc<argvsize)) {
		if ((strlen(de->d_name) > 3) && !strcasecmp(de->d_name + strlen(de->d_name) - 4, ".mp3")) {
			strncpy(fns[files], de->d_name, sizeof(fns[files]) - 1);
			argv[argc++] = fns[files];
			files++;
		}
	}
	argv[argc] = NULL;
	closedir(dir);

#if EXTRA_LOG
	fprintf(stderr,"%d files total, %d args total\n", files, argc);
	{
		int x;
		for (x=0;argv[x];x++)
			fprintf(stderr,"arg%d: %s\n", x, argv[x]);
	}
#endif	
	if (!files) {
		ast_log(LOG_WARNING, "Found no files in '%s'\n", _this->dir);
		return -1;
	}

	return 0;
}

static int mohclass_is_equal( struct mohclass* _this, struct mohclass* _that ) {
	if( strcasecmp(_this->class,_that->class) ) return 0;
	if( strcmp(_this->dir,_that->dir) ) return 0;
	if( _this->quiet!=_that->quiet ) return 0;
	if( _this->single!=_that->single ) return 0;
	if( _this->custom!=_that->custom ) return 0;
	return 1;
}

static int mohclass_spawn_mp3( struct mohclass* _this ) {
	int fds[2];
	char *argv[MAX_MP3S + 50];

	if( !_this ) 
		return -1;

	if( mohclass_make_mp3_args(_this,ELEMOF(argv),argv) ) {
		ast_log(LOG_WARNING,"Cannot make mp3 arguments\n");
		return -1;
	}

	if (pipe(fds)) {	
		ast_log(LOG_WARNING, "Pipe failed\n");
		return -1;
	}
	_this->pid = fork();
	if (_this->pid < 0) {
		close(fds[0]);
		close(fds[1]);
		ast_log(LOG_WARNING, "Fork failed: %s\n", strerror(errno));
		return -1;
	}
	if (!_this->pid) {
		int x;
		close(fds[0]);
		/* Stdout goes to pipe */
		dup2(fds[1], STDOUT_FILENO);
		/* Close unused file descriptors */
		for (x=3;x<8192;x++)
			close(x);
		/* Child */
		chdir(_this->dir);
		if(_this->custom) {
			execv(argv[0], argv);
		} else {
			/* Default install is /usr/local/bin */
			execv(LOCAL_MPG_123, argv);
			/* Many places have it in /usr/bin */
			execv(MPG_123, argv);
			/* Check PATH as a last-ditch effort */
			execvp("mpg123", argv);
		}
		ast_log(LOG_WARNING, "Exec failed: %s\n", strerror(errno));
		close(fds[1]);
		exit(1);
	} else {
		/* Parent */
		close(fds[1]);
	}
	return fds[0];
}

static int mohclass_kill_mp3( struct mohclass* _this ) {
	int result = 0;
	if( _this->pid ) {
		ast_log(LOG_DEBUG, "killing %d!\n", _this->pid);
		if( kill(_this->pid,SIGKILL) && (errno!=ESRCH) ) {
			ast_log(LOG_WARNING,"Unable to kill moh process, %d\n",errno);
			result++;
		}
		else {
			time_t stime = time(NULL);
			char buff[8192];
			int bytes, tbytes=0;
			while( (bytes=read(_this->srcfd, buff, sizeof(buff))) && (time(NULL)<(stime+5)) ) {
				tbytes += tbytes;
			}
			ast_log(LOG_DEBUG, "mpg123 pid %d and child died after %d bytes read\n", _this->pid, tbytes);
			close(_this->srcfd);
			_this->srcfd = -1;
			_this->pid = 0;
		}
	}
	else {
		ast_log(LOG_WARNING,"Attempt to kill a not running mp3 process for moh class '%s'\n",_this->class);
	}
	return result;
}

static void mohclass_set_destroyme( struct mohclass* _this, int destroyme ) {
	ast_mutex_lock(&_this->lock);
	_this->destroyme = destroyme;
	ast_mutex_unlock(&_this->lock);
}

static int mohclass_get_destroyme( struct mohclass* _this ) {
	int destroyme;
	ast_mutex_lock(&_this->lock);
	destroyme = _this->destroyme;
	ast_mutex_unlock(&_this->lock);
	return destroyme;
}

static void* mohclass_monmp3thread(void *data)
{
#define	MOH_MS_INTERVAL		100

	struct mohclass *_this = data;
	struct mohdata *moh;
	short sbuf[8192];
	int res, res2;
	struct timeval tv;
	struct timeval tv_tmp;
	long error_sec, error_usec;
	long delay;
	tv_tmp.tv_sec = 0;
	tv_tmp.tv_usec = 0;
	tv.tv_sec = 0;
	tv.tv_usec = 0;
	error_sec = 0;
	error_usec = 0;
	while( !mohclass_get_destroyme(_this) ) {
		/* Spawn mp3 player if it's not there */
		if (_this->srcfd < 0) {
			if ((_this->srcfd = mohclass_spawn_mp3(_this)) < 0) {
				ast_log(LOG_WARNING, "unable to spawn mp3player\n");
				/* Try again later */
				sleep(500);
			}
		}
#if ZAPATA_MOH
		if (_this->pseudofd > -1) {
			/* Pause some amount of time */
			char buf[8192];
			res = read(_this->pseudofd, buf, sizeof(buf));
		} else {
#endif
			/* Reliable sleep */
			if (gettimeofday(&tv_tmp, NULL) < 0) {
				ast_log(LOG_NOTICE, "gettimeofday() failed!\n");
				return NULL;
			}
			if (((unsigned long)(tv.tv_sec) > 0)&&((unsigned long)(tv.tv_usec) > 0)) {
				if ((unsigned long)(tv_tmp.tv_usec) < (unsigned long)(tv.tv_usec)) {
					tv_tmp.tv_usec += 1000000;
					tv_tmp.tv_sec -= 1;
				}
				error_sec = (unsigned long)(tv_tmp.tv_sec) - (unsigned long)(tv.tv_sec);
				error_usec = (unsigned long)(tv_tmp.tv_usec) - (unsigned long)(tv.tv_usec);
			} else {
				error_sec = 0;
				error_usec = 0;
			}
			if (error_sec * 1000 + error_usec / 1000 < MOH_MS_INTERVAL) {
				tv.tv_sec = tv_tmp.tv_sec + (MOH_MS_INTERVAL/1000 - error_sec);
				tv.tv_usec = tv_tmp.tv_usec + ((MOH_MS_INTERVAL % 1000) * 1000 - error_usec);
				delay = (MOH_MS_INTERVAL/1000 - error_sec) * 1000 +
							((MOH_MS_INTERVAL % 1000) * 1000 - error_usec) / 1000;
			} else {
				ast_log(LOG_NOTICE, "Request to schedule in the past?!?!\n");
				tv.tv_sec = tv_tmp.tv_sec;
				tv.tv_usec = tv_tmp.tv_usec;
				delay = 0;
			}
			if (tv.tv_usec > 1000000) {
				tv.tv_sec++;
				tv.tv_usec-= 1000000;
			}
			if (delay > 0)
				usleep(delay * 1000);
			res = 800;		/* 800 samples */
#ifdef ZAPATA_MOH
		}
#endif
		if (!_this->members)
			continue;
		/* Read mp3 audio */
		if ((res2 = read(_this->srcfd, sbuf, res * 2)) != res * 2) {
			if (!res2)
				mohclass_kill_mp3(_this);
			else
				ast_log(LOG_DEBUG, "Read %d bytes of audio while expecting %d\n", res2, res * 2);
			continue;
		}
		ast_mutex_lock(&_this->lock);
		moh = _this->members;
		while(moh) {
			/* Write data */
			if ((res = write(moh->pipe[1], sbuf, res2)) != res2) {
				if( errno==EAGAIN ) {
					/* the reader thread is too busy */
				}
				else {
#if EXTRA_LOG
					fprintf(stderr,"mohclass_monmp3thread:unable to write to pipe: res=%d,res2=%d,errno=%d\n",res,res2,errno);
#endif
					if (option_debug)
						ast_log(LOG_DEBUG, "Only wrote %d of %d bytes to pipe\n", res, res2);
				}
			}
			moh = moh->next;
		}
		ast_mutex_unlock(&_this->lock);
	}
	mohclass_kill_mp3(_this);
	return NULL;
}

static int mohclass_start( struct mohclass* _this ) {
	int result = -1;
	if( _this->thread==0 ) {
		result = ast_pthread_create(&_this->thread,NULL,mohclass_monmp3thread,_this);
	}
	else {
		ast_log(LOG_WARNING,"An attempt to start a running thread for moh class '%s'\n",_this->class);
	}
	return result;
	
}

static int mohclass_stop( struct mohclass* _this ) {
	int result = 0;
	if( _this->thread==0 ) {
		ast_log(LOG_WARNING,"An attempt to stop a not running thread for moh class '%s'\n",_this->class);
		result++;
	}
	else {
		mohclass_set_destroyme(_this,1);
		if( pthread_join(_this->thread,NULL) ) {
			ast_log(LOG_WARNING,"Unable to join moh thread, %d\n",errno);
			result++;
		}
		else {
			_this->thread = 0;
		}
	}
	return result;
}

static struct mohclass* mohclass_delete( struct mohclass* _this ) {
	struct mohclass* result = _this->next;
	mohclass_stop(_this);
	ast_mutex_destroy(&_this->lock);
	free(_this);
	return result;
}

#ifdef POSTGRES_MOH
/*********************************************************************************
 * changes to integrate MOH with postgres database
 ********************************************************************************/
static PGconn* _dbconn;
static char    _dbget_mohclasses_query[512];

static int sql_connect(void) {
	struct ast_config *cfg;
	struct ast_variable *var;
	cfg = ast_load("musiconhold.conf");
	if (cfg) {
		PGconn* dbconn;
		char    dboption[512] = {0};
		char    dbget_mohclasses_query[sizeof(_dbget_mohclasses_query)] = {0};
		var = ast_variable_browse(cfg,"general");
		while(var) {
			if( !strcasecmp(var->name,"dboption") ) {
				strncpy(dboption,var->value,sizeof(dboption)-1);
			}
			else if( !strcasecmp(var->name,"get_mohclasses_query") ) {
				strncpy(dbget_mohclasses_query,var->value,sizeof(dbget_mohclasses_query)-1);
			}
			var = var->next;
		}
		/* try to open the database coonection */
		if( (dbconn=PQconnectdb(dboption))!=NULL ) {
			if( PQstatus(dbconn)==CONNECTION_OK ) {
				_dbconn = dbconn;
				strcpy(_dbget_mohclasses_query,dbget_mohclasses_query);
			}
			else {
				/* Danger!!!! the database password may appear in the logs here */
				ast_log(LOG_WARNING,"Unable to open database connection to '%s'\n",dboption);
				PQfinish(dbconn);
			}
		}
		ast_destroy(cfg);
	}
	else {
		ast_log(LOG_WARNING,"Unable to read musiconhold.conf\n");
	}
	return _dbconn ? 0 : -1;
}

static void sql_disconnect(void) {
	if( _dbconn ) {
		PQfinish(_dbconn);
		_dbconn = NULL;
	}
}

static int read_moh_classes( struct mohclass** pmoh_classes ) {
	int result = -1;
	PGresult* pgresult;
	if( (pgresult=PQexec(_dbconn,_dbget_mohclasses_query))!=NULL ) {
		ExecStatusType pgstatus = PQresultStatus(pgresult);
		if( (pgstatus!=PGRES_BAD_RESPONSE) && (pgstatus!=PGRES_NONFATAL_ERROR) && (pgstatus!=PGRES_FATAL_ERROR) ) {
			int nameColumn     = PQfnumber(pgresult,"name");
			int modeColumn     = PQfnumber(pgresult,"mode");
			int dirColumn      = PQfnumber(pgresult,"dir");
			int miscargsColumn = PQfnumber(pgresult,"miscargs");
			if( (nameColumn>=0) && (modeColumn>=0) && (dirColumn>=0) && (miscargsColumn>=0) ) {
				int n;
				int rows               = PQntuples(pgresult);
				int errors_so_far      = 0;
				int errors_to_tolerate = 2;
				struct mohclass** presult = pmoh_classes;
				*pmoh_classes = NULL;
				for( n=0; n<rows; n++ ) {
					*presult = mohclass_new(
						PQgetisnull(pgresult,n,nameColumn)     ? NULL : PQgetvalue(pgresult,n,nameColumn),
						PQgetisnull(pgresult,n,modeColumn)     ? NULL : PQgetvalue(pgresult,n,modeColumn),
						PQgetisnull(pgresult,n,dirColumn)      ? NULL : PQgetvalue(pgresult,n,dirColumn),
						PQgetisnull(pgresult,n,miscargsColumn) ? NULL : PQgetvalue(pgresult,n,miscargsColumn));
					if( *presult ) {
						(*presult)->next = 0;
						presult = &((*presult)->next);
					}
					else if( ++errors_so_far>errors_to_tolerate ) {
						ast_log(LOG_WARNING,"Musiconhold database misconfiguration detected. Giving up\n");
						break;
					}
					else {
						ast_log(LOG_WARNING,"Cannot read moh class\n");
					}
				}
				if( n==rows ) {
					result = 0;
				}
				else {
					while( *pmoh_classes ) {
						*pmoh_classes = mohclass_delete(*pmoh_classes);
					}
				}
			}
			else {
				ast_log(LOG_WARNING,"Required columns are missing\n");
			}
		}
		else {
			ast_log(LOG_WARNING,"%s\n",PQcmdStatus(pgresult));
		}
		PQclear(pgresult);
	}
	else {
		ast_log(LOG_WARNING,"SQL query '%s' returned NULL\n",_dbget_mohclasses_query);
	}
	return result;
}

static int sync_moh_classes(void) {
	int result;
	struct mohclass* old_mohclasses = mohclasses; /* This is the global variable */
	struct mohclass* new_mohclasses;
	if( (result=read_moh_classes(&new_mohclasses))==0 ) {

		struct mohclass*  new_mohclass;
		struct mohclass*  old_mohclass;
		struct mohclass*  prev_new_mohclass = NULL;

		/* mark the old classes as dead (they likely are to be replaced) */
		for( old_mohclass=old_mohclasses; old_mohclass; old_mohclass=old_mohclass->next ) {
			old_mohclass->deadinsyncing = 1;
		}

		/* start the new moh classes */
		for( new_mohclass=new_mohclasses; new_mohclass; ) {
			old_mohclass = mohclass_find(old_mohclasses,new_mohclass->class);
			if( !old_mohclass && !mohclass_start(new_mohclass) ) {
#if EXTRA_LOG
				fprintf(stderr,"Started new moh class %s\n",new_mohclass->class);
#endif
				prev_new_mohclass = new_mohclass;
				new_mohclass = new_mohclass->next;
			}
			else if( !mohclass_is_equal(old_mohclass,new_mohclass) && !mohclass_start(new_mohclass) ) {
#if EXTRA_LOG
				fprintf(stderr,"Started changed moh class %s\n",new_mohclass->class);
#endif
				old_mohclass->new_mohclass = new_mohclass; /* will pass the members over */
				prev_new_mohclass = new_mohclass;
				new_mohclass = new_mohclass->next;
			}
			else {
#if EXTRA_LOG
				fprintf(stderr,"Ignoring new class '%s' (it is probably equal)\n",new_mohclass->class);
#endif				
				old_mohclass->deadinsyncing = 0;
				if( prev_new_mohclass )
					prev_new_mohclass->next = new_mohclass->next;
				else
					new_mohclasses = new_mohclass->next;
				new_mohclass = mohclass_delete(new_mohclass);
			}
		}

		/* destroy all dead old moh classes, Living classes merge into the new classes */
		ast_mutex_lock(&moh_lock);
		for( old_mohclass=old_mohclasses; old_mohclass; ) {
			if( old_mohclass->deadinsyncing ) {
				/* destroy the old class, possibly passing members over */
				if( old_mohclass->new_mohclass ) {
					ast_mutex_lock(&old_mohclass->lock);
					old_mohclass->new_mohclass->members = old_mohclass->members;
					old_mohclass->members = 0;
					ast_mutex_unlock(&old_mohclass->lock);
				}
				old_mohclass = mohclass_delete(old_mohclass);
			}
			else {
				/* stick it into the beginning of new_mohclasses */
				struct mohclass* tmp = old_mohclass->next;
				old_mohclass->next = new_mohclasses;
				new_mohclasses = old_mohclass;
				old_mohclass = tmp;
			}
		}
		mohclasses = new_mohclasses;
		ast_mutex_unlock(&moh_lock);
	}
	else {
		/* a proper error message is already logged */
	}
	return result;
}

#endif

static struct mohclass *get_mohbyname(const char *name)
{
#ifdef POSTGRES_MOH
	struct mohclass* result;
	if( !(result=mohclass_find(mohclasses,name)) ) {
		/* sync and try again */
		if( !sync_moh_classes() ) {
			return mohclass_find(mohclasses,name);
		}
	}
	return result;
#else	
	return mohclass_find(mohclasses,name);
#endif
}

#ifndef POSTGRES_MOH
static int moh_register(char *classname, char *mode, char *param, char *miscargs)
{
	struct mohclass *mclass;
	ast_mutex_lock(&moh_lock);
	mclass = get_mohbyname(classname);
	ast_mutex_unlock(&moh_lock);
	if (mclass) {
		ast_log(LOG_WARNING, "Music on Hold '%s' already exists\n", classname);
		return -1;
	}
	mclass = mohclass_new(classname,mode,param,miscargs);
	if (!mclass)
		return -1;
	if (mohclass_start(mclass)) {
		ast_log(LOG_WARNING, "Unable to start moh class...\n");
		mohclass_delete(mclass);
		return -1;
	}
	ast_mutex_lock(&moh_lock);
	mclass->next = mohclasses;
	mohclasses = mclass;
	ast_mutex_unlock(&moh_lock);
	return 0;
}
#endif /* POSTGRES_MOH */

/*********************************************************************************
 *
 ********************************************************************************/
static int moh0_exec(struct ast_channel *chan, void *data)
{
	if (ast_moh_start(chan, data)) {
		ast_log(LOG_WARNING, "Unable to start music on hold (class '%s') on channel %s\n", (char *)data, chan->name);
		return -1;
	}
	while(!ast_safe_sleep(chan, 10000));
	return -1;
}

static int moh1_exec(struct ast_channel *chan, void *data)
{
	int res;
	if (!data || !atoi(data)) {
		ast_log(LOG_WARNING, "WaitMusicOnHold requires an argument (number of seconds to wait)\n");
		return -1;
	}
	if (ast_moh_start(chan, NULL)) {
		ast_log(LOG_WARNING, "Unable to start music on hold (class '%s') on channel %s\n", (char *)data, chan->name);
		return -1;
	}
	res = ast_safe_sleep(chan, atoi(data) * 1000);
	ast_moh_stop(chan);
	return res;
}

static int moh2_exec(struct ast_channel *chan, void *data)
{
	if (!data || ast_strlen_zero(data)) {
		ast_log(LOG_WARNING, "SetMusicOnHold requires an argument (class)\n");
		return -1;
	}
	strncpy(chan->musicclass, data, sizeof(chan->musicclass) - 1);
	return 0;
}

static struct mohdata *mohalloc(struct mohclass *cl)
{
	struct mohdata *moh;
	long flags;
	moh = malloc(sizeof(struct mohdata));
	if (!moh)
		return NULL;
	memset(moh, 0, sizeof(struct mohdata));
	if (pipe(moh->pipe)) {
		ast_log(LOG_WARNING, "Failed to create pipe: %s\n", strerror(errno));
		free(moh);
		return NULL;
	}
	/* Make entirely non-blocking */
	flags = fcntl(moh->pipe[0], F_GETFL);
	fcntl(moh->pipe[0], F_SETFL, flags | O_NONBLOCK);
	flags = fcntl(moh->pipe[1], F_GETFL);
	fcntl(moh->pipe[1], F_SETFL, flags | O_NONBLOCK);
	moh->parent = cl;
	moh->next = cl->members;
	ast_mutex_lock(&cl->lock);
	cl->members = moh;
	ast_mutex_unlock(&cl->lock);
	return moh;
}

static void moh_release(struct ast_channel *chan, void *data)
{
	struct mohdata *moh = data, *prev, *cur;
	int oldwfmt;
	ast_mutex_lock(&moh->parent->lock);
	/* Unlink */
	prev = NULL;
	cur = moh->parent->members;
	while(cur) {
		if (cur == moh) {
			if (prev)
				prev->next = cur->next;
			else
				moh->parent->members = cur->next;
			break;
		}
		prev = cur;
		cur = cur->next;
	}
	ast_mutex_unlock(&moh->parent->lock);
	close(moh->pipe[0]);
	close(moh->pipe[1]);
	oldwfmt = moh->origwfmt;
	free(moh);
	if (chan) {
		if (oldwfmt && ast_set_write_format(chan, oldwfmt)) 
			ast_log(LOG_WARNING, "Unable to restore channel '%s' to format %s\n", chan->name, ast_getformatname(oldwfmt));
		if (option_verbose > 2)
			ast_verbose(VERBOSE_PREFIX_3 "Stopped music on hold on %s\n", chan->name);
	}
}

static void *moh_alloc(struct ast_channel *chan, void *params)
{
	struct mohdata *res;
	struct mohclass *class;
	ast_mutex_lock(&moh_lock);
	class = get_mohbyname(params);
	if (class)
		res = mohalloc(class);
	else {
		if (strcasecmp(params, "default"))
			ast_log(LOG_WARNING, "No class: %s\n", (char *)params);
		res = NULL;
	}
	ast_mutex_unlock(&moh_lock);
	if (res) {
		res->origwfmt = chan->writeformat;
		if (ast_set_write_format(chan, AST_FORMAT_SLINEAR)) {
			ast_log(LOG_WARNING, "Unable to set '%s' to signed linear format\n", chan->name);
			moh_release(NULL, res);
			res = NULL;
		}
#if 0
		/* Allow writes to interrupt */
		chan->writeinterrupt = 1;
#endif		
		if (option_verbose > 2)
			ast_verbose(VERBOSE_PREFIX_3 "Started music on hold, class '%s', on %s\n", (char *)params, chan->name);
	}
	return res;
}

static int moh_generate(struct ast_channel *chan, void *data, int len, int samples)
{
	struct ast_frame f;
	struct mohdata *moh = data;
	short buf[1280 + AST_FRIENDLY_OFFSET / 2];
	int res;

	len = samples * 2;
	if (len > sizeof(buf) - AST_FRIENDLY_OFFSET) {
		ast_log(LOG_WARNING, "Only doing %d of %d requested bytes on %s\n", (int)sizeof(buf), (int)len, chan->name);
		len = sizeof(buf) - AST_FRIENDLY_OFFSET;
	}
	res = read(moh->pipe[0], buf + AST_FRIENDLY_OFFSET/2, len);
#if 0
	if (res != len) {
		ast_log(LOG_WARNING, "Read only %d of %d bytes: %s\n", res, len, strerror(errno));
	}
#endif
	if (res > 0) {
		memset(&f, 0, sizeof(f));
		f.frametype = AST_FRAME_VOICE;
		f.subclass = AST_FORMAT_SLINEAR;
		f.mallocd = 0;
		f.datalen = res;
		f.samples = res / 2;
		f.data = buf + AST_FRIENDLY_OFFSET / 2;
		f.offset = AST_FRIENDLY_OFFSET;
		if (ast_write(chan, &f)< 0) {
			ast_log(LOG_WARNING, "Failed to write frame to '%s': %s\n", chan->name, strerror(errno));
			return -1;
		}
#if EXTRA_LOG>1
		else {
			fprintf(stderr,"moh_generate:Wrote %d bytes to channel\n",res);
		}
#endif
	}
#if EXTRA_LOG
	else if( res<0 ) {
		if( errno==EAGAIN ) {
			/* the writer thread hasn't written anything yet */
		}
		else {
			fprintf(stderr,"moh_generate: unable to read from pipe,res=%d,errno=%d\n",res,errno);
		}
	}
#endif
	return 0;
}

static struct ast_generator mohgen = 
{
	alloc: moh_alloc,
	release: moh_release,
	generate: moh_generate,
};

int ast_moh_start(struct ast_channel *chan, char *class)
{
	if (!class || ast_strlen_zero(class))
		class = chan->musicclass;
	if (!class || ast_strlen_zero(class))
		class = "default";
	return ast_activate_generator(chan, &mohgen, class);
}

void ast_moh_stop(struct ast_channel *chan)
{
	ast_deactivate_generator(chan);
}

static int load_moh_classes(void)
{
#ifdef POSTGRES_MOH
	/* Do not really have to start moh classes here - this can be done in a 
	   lazy fashion when the first get_mohbyname is called */
	return sql_connect();
#else
	struct ast_config *cfg;
	struct ast_variable *var;
	char *data;
	char *args;
	cfg = ast_load("musiconhold.conf");
	if (cfg) {
		var = ast_variable_browse(cfg, "classes");
		while(var) {
			data = strchr(var->value, ':');
			if (data) {
				*data = '\0';
				data++;
				args = strchr(data, ',');
				if (args) {
					*args = '\0';
					args++;
				}
				if( moh_register(var->name, var->value, data,args)!=0 ) {
					ast_destroy(cfg);
					return -1;
				}
			}
			var = var->next;
		}
		ast_destroy(cfg);
	}
	return 0;
#endif
}

static void ast_moh_destroy(void)
{
	if (option_verbose > 1)
		ast_verbose(VERBOSE_PREFIX_2 "Destroying any remaining musiconhold processes\n");
	ast_mutex_lock(&moh_lock);
	while( mohclasses ) {
		mohclasses = mohclass_delete(mohclasses);
	}
	ast_mutex_unlock(&moh_lock);
#ifdef POSTGRES_MOH
	sql_disconnect();
#endif
}

int load_module(void)
{
	int res = load_moh_classes();
	if(!res) {
		ast_register_atexit(ast_moh_destroy);
		res = ast_register_application(app0, moh0_exec, synopsis0, descrip0);
		if (!res) {
			res = ast_register_application(app1, moh1_exec, synopsis1, descrip1);
			if (!res)
				res = ast_register_application(app2, moh2_exec, synopsis2, descrip2);
		}
	}
	return res;
}

int unload_module(void)
{
	return -1;
}

char *description(void)
{
	return "Music On Hold Resource";
}

int usecount(void)
{
	/* Never allow Music On Hold to be unloaded
	   unresolve needed symbols in the dialer */
#if 0
	int res;
	STANDARD_USECOUNT(res);
	return res;
#else
	return 1;
#endif
}

char *key()
{
	return ASTERISK_GPL_KEY;
}
